@@ -1,4 +1,4 @@
-####-*- mode: R; kept-old-versions: 12;  kept-new-versions: 20; -*-
+####-*- mode: R; kept-old-versions: 15;  kept-new-versions: 30; -*-
 
 ### Original file:
 ### copyright 1991 Department of Statistics, Univeristy of Washington
@@ -6,8 +6,54 @@
 ### Patched by Friedrich.Leisch, for use with R, 22.1.1997
 ### fixed & changed by Martin Maechler, since Dec 2003
 
-".First.lib" <- function(lib, pkg) library.dynam("fracdiff", pkg, lib)
+## no longer, with useDynLib() in ../NAMESPACE
+## no ".First.lib" <- function(lib, pkg) library.dynam("fracdiff", pkg, lib)
 
+p0 <- function(...) paste(..., sep="")
+
+.fdcov <- function(x, d, h, nar, nma, hess, fdf.work)
+{
+    npq <- as.integer(nar + nma)
+    npq1 <- npq + 1L # integer, too
+    stopifnot(length(di <- dim(hess)) == 2, di == c(npq1, npq1))
+    fdc <- .C("fdcov", ## --> ../src/fdhess.c
+	      x,
+	      d,
+	      h = as.double(if(missing(h)) -1 else h),
+	      hd = double(npq1),
+	      cov = hess, npq1,
+	      cor = hess, npq1,
+	      se = double(npq1),
+	      fdf.work,
+	      info = integer(1),
+	      PACKAGE = "fracdiff")[c("h","hd", "cov","cor", "se", "info")]
+
+    f.msg <-
+	if(fdc$info) {
+	    msg <-
+		switch(fdc$info,
+		       "fdcov problem in gamma function",	# 1
+		       "singular Hessian",			# 2
+		       ## FIXME improve: different reasons for info = 3 :
+		       "unable to compute correlation matrix; maybe change 'h'",
+								# 3
+		       stop("error in gamma function"))		# 4
+	    warning(msg, call. = FALSE)
+	    msg
+	} else "ok"
+    se.ok <- fdc$info %in% 0:2
+    nam <- "d"
+    if(nar) nam <- c(nam, p0("ar", 1:nar))
+    if(nma) nam <- c(nam, p0("ma", 1:nma))
+
+    dimnames(fdc$cov) <- dn <- list(nam, nam)
+    if(se.ok) dimnames(fdc$cor) <- dn
+    list(msg = f.msg, d = d, nam = nam,
+	 h = fdc$h, hd = fdc$hd, se.ok = se.ok,
+	 covariance.dpq = fdc$cov,
+	 stderror.dpq	= if(se.ok) fdc$se, # else NULL
+	 correlation.dpq= if(se.ok) fdc$cor)
+}## end{.fdcov}
 
 fracdiff <- function(x, nar = 0, nma = 0,
                      ar = rep(NA, max(nar, 1)), ma = rep(NA, max(nma, 1)),
@@ -39,9 +85,9 @@ fracdiff <- function(x, nar = 0, nma = 0,
         stop("multivariate time series not allowed")
     n <- length(x)
     if(round(nar) != nar || nar < 0 || round(nma) != nma || nma < 0)
-        stop("'nar' and 'nma' must be non-negative integers")
+        stop("'nar' and 'nma' must be non-negative integer numbers")
     npq <- as.integer(nar + nma)
-    npq1 <- as.integer(npq + 1:1)
+    npq1 <- npq + 1L # integer, too
     lenw <- max(npq + 2*(n + M),
                 3*n + (n+6)*npq + npq %/% 2 + 1,
                 31 * 12, ## << added because checked in ../src/fdcore.f
@@ -55,89 +101,82 @@ fracdiff <- function(x, nar = 0, nma = 0,
     x <- as.double(x)
 
     ## this also initializes "common blocks" that are used in .C(.) calls :
-    result <- .C("fracdf",
-		 x,
-		 n,
-		 as.integer(M),
-		 as.integer(nar),
-		 as.integer(nma),
-		 dtol = as.double(dtol),
-		 drange = as.double(drange),
-		 hood = double(1),
-		 d = double(1),
-		 ar = as.double(ar),
-		 ma = as.double(ma),
-		 w = double(lenw),
-		 lenw = lenw,
-		 iw = integer(npq), ## <<< new int-work array
-		 info = integer(1),
-		 .Machine$double.xmin,
-		 .Machine$double.xmax,
-		 .Machine$double.neg.eps,
-		 .Machine$double.eps,
-		 PACKAGE = "fracdiff")
-    if(result$info)
-        switch(result$info,
-               stop("insufficient workspace; need ", result$lenw,
-                    " instead of just ", lenw),
-               stop("error in gamma function"),
-               stop("invalid MINPACK input"),
-               warning("warning in gamma function"),
-               warning("optimization failure"),
-               warning("optimization limit reached"))
+    fdf <- .C("fracdf",
+	      x,
+	      n,
+	      as.integer(M),
+	      as.integer(nar),
+	      as.integer(nma),
+	      dtol = as.double(dtol),
+	      drange = as.double(drange),
+	      hood = double(1),
+	      d = double(1),
+	      ar = as.double(ar),
+	      ma = as.double(ma),
+	      w = double(lenw),
+	      lenw = lenw,
+	      iw = integer(npq), ## <<< new int-work array
+	      info = integer(1),
+	      .Machine$double.xmin,
+	      .Machine$double.xmax,
+	      .Machine$double.neg.eps,
+	      .Machine$double.eps,
+	      PACKAGE = "fracdiff")[c("dtol","drange","hood",
+	      "d", "ar", "ma", "w", "lenw", "info")]
+
+    fd.msg <-
+        if(fdf$info) {
+	    msg <-
+                switch(fdf$info,
+                       stop("insufficient workspace; need ", fdf$lenw,
+                            " instead of just ", lenw),		# 1
+                       stop("error in gamma function"),		# 2
+                       stop("invalid MINPACK input"),		# 3
+                       "warning in gamma function",		# 4
+                       "C fracdf() optimization failure",	# 5
+                       "C fracdf() optimization limit reached") # 6
+            ## otherwise
+            ## stop("unknown .C(fracdf, *) info -- should not happen")
+	    warning(msg, call. = FALSE, immediate. = TRUE)
+	    msg
+	} else "ok"
+
+    if(nar == 0) fdf$ar <- numeric(0)
+    if(nma == 0) fdf$ma <- numeric(0)
 
     hess <- .C("fdhpq",
-               ## x,
-               hess = double(npq1 * npq1),
+               hess = matrix(double(1), npq1, npq1),
                npq1,
-               result$w,
+               fdf$w,
                PACKAGE = "fracdiff")$hess
 
-    temp <- .C("fdcov",
-               x,
-               result$d,
-               h = as.double(if(missing(h)) -1 else h),
-               hd = double(npq1),
-               cov = hess, npq1,
-               cor = hess, npq1,
-               se = double(npq1),
-               result$w,
-               info = integer(1),
-               PACKAGE = "fracdiff")
-    if(temp$info) {
-        msg <-
-            switch(temp$info,
-                   "warning in gamma function",
-                   "singular Hessian",
-                   "unable to compute correlation matrix",
-                   stop("error in gamma function"))
-        warning(msg)
-	result$msg <- msg
-    } else {
-	result$msg <- "ok"
-    }
-    se.ok <- temp$info != 0 || temp$info < 3 ## FIXME -- illogical!!
-    if(npq == 0) {
-        result$ar <- NULL
-        result$ma <- NULL
-    }
-    nam <- "d"
-    if(nar) nam <- c(nam, paste("ar", 1:nar, sep = ""))
-    if(nma) nam <- c(nam, paste("ma", 1:nma, sep = ""))
-    hess <- matrix(hess, nrow = npq1, ncol = npq1, dimnames = list(nam, nam))
-    hess[1, ] <- temp$hd
+    ## NOTA BENE: The above  hess[.,.]  is further "transformed",
+    ##            well, added to  and inverted  in fdcov :
+    ## Cov == (-H)^{-1} == solve(-H)
+
+    ## Note that the following can be "redone" using fracdiff.var() :
+
+    fdc <- .fdcov(x, fdf$d, h,
+                  nar=nar, nma=nma, hess=hess, fdf.work = fdf$w)
+
+    dimnames(hess) <- dimnames(fdc$covariance.dpq)
+    hess[1, ] <- fdc$hd
     hess[row(hess) > col(hess)] <- hess[row(hess) < col(hess)]
-    structure(list(log.likelihood = result$hood, n = n, msg = result$msg,
-		   d = result$d, ar = result$ar, ma = result$ma,
-		   covariance.dpq = array(temp$cov, c(npq1,npq1), list(nam,nam)),
-		   stderror.dpq	  = if(se.ok) temp$se, # else NULL
-		   correlation.dpq= if(se.ok) array(temp$cor, c(npq1, npq1)),
-		   h = temp$h, d.tol = result$dtol, M = M, hessian.dpq = hess,
+
+    structure(list(log.likelihood = fdf$hood, n = n,
+		   msg = c(fracdf = fd.msg, fdcov = fdc$msg),
+		   d = fdf$d, ar = fdf$ar, ma = fdf$ma,
+		   covariance.dpq = fdc$covariance.dpq,
+		   stderror.dpq	  = if(fdc$se.ok) fdc$stderror.dpq, # else NULL
+		   correlation.dpq= if(fdc$se.ok) fdc$correlation.dpq,
+		   h = fdc$h, d.tol = fdf$dtol, M = M, hessian.dpq = hess,
 		   length.w = lenw, call = cl),
 	      class = "fracdiff")
 }
 
-### FIXME [modularity]: a lot of this is "cut & paste" also in fracdiff() itselff
+### FIXME [modularity]: a lot of this is "cut & paste" also in fracdiff() itself
+### ----- NOTABLY, now use  .fdcov() !
+
 fracdiff.var <- function(x, fracdiff.out, h)
 {
     if(!is.numeric(h))
@@ -165,7 +204,7 @@ fracdiff.var <- function(x, fracdiff.out, h)
        .Machine$double.eps,
        PACKAGE = "fracdiff")
     ## Re compute Covariance Matrix:
-    temp <- .C("fdcov",
+    fdc <- .C("fdcov",
                as.double(x),
                as.double(fracdiff.out$d),
                h = as.double(h),
@@ -180,31 +219,29 @@ fracdiff.var <- function(x, fracdiff.out, h)
                            rep(0, lwork))),
                info = integer(1),
                PACKAGE = "fracdiff")
-    if(temp$info) {
-	msg <-
-	    switch(temp$info,
-		   "warning in gamma function",
-		   "singular Hessian",
-		   "unable to compute correlation matrix",
-		   stop("error in gamma function"))
-	warning(msg)
-	fracdiff.out$msg <- msg
-    } else {
-	fracdiff.out$msg <- "ok"
-    }
-    se.ok <- temp$info != 0 || temp$info < 3 ## << FIXME -- illogical!!
+## FIXME: should be *automatically* same messages as inside fracdiff() above!
+    fracdiff.out$msg <-
+        if(fdc$info) {
+            msg <-
+                switch(fdc$info,
+                       "warning in gamma function",
+                       "singular Hessian",
+                       "unable to compute correlation matrix",
+                       stop("error in gamma function"))
+            warning(msg)
+        } else "ok"
+    se.ok <- fdc$info != 0 || fdc$info < 3 ## << FIXME -- illogical!!
     nam <- "d"
-    if(p) nam <- c(nam, paste("ar", 1:p, sep = ""))
-    if(q) nam <- c(nam, paste("ma", 1:q, sep = ""))
-    fracdiff.out$h <- temp$h
-    fracdiff.out$covariance.dpq <-
-        matrix(temp$cov, nrow = npq1, ncol = npq1, dimnames = list(nam, nam))
-    fracdiff.out$stderror.dpq    <- if(se.ok) temp$se # else NULL
-    fracdiff.out$correlation.dpq <- if(se.ok) array(temp$cor, c(npq1, npq1))
-    fracdiff.out$hessian.dpq[1, ] <- temp$hd
-    fracdiff.out$hessian.dpq[, 1] <- temp$hd
+    if(p) nam <- c(nam, p0("ar", 1:p))
+    if(q) nam <- c(nam, p0("ma", 1:q))
+    fracdiff.out$h <- fdc$h
+    fracdiff.out$covariance.dpq <- array(fdc$cov, c(npq1,npq1), list(nam,nam))
+    fracdiff.out$stderror.dpq    <- if(se.ok) fdc$se # else NULL
+    fracdiff.out$correlation.dpq <- if(se.ok) array(fdc$cor, c(npq1, npq1))
+    fracdiff.out$hessian.dpq[1, ] <- fdc$hd
+    fracdiff.out$hessian.dpq[, 1] <- fdc$hd
     fracdiff.out
-}
+}## end{ fracdiff.var() }
 
 ## MM:  Added things for more  arima.sim() compatibility.
 ##      really, 'mu' is nonsense since can be done separately (or via 'innov').
